use std::{str::Chars, sync::Arc};

pub(crate) use self::{
    error::{LexerError, LexerErrorKind},
    symbol::Symbol,
    token::{Token, TokenKind},
};
use crate::span::{Source, Span};

mod error;
mod symbol;
mod token;

/// Determine if the current character is a separator, performing 1-character
/// lookahead as needed to handle multi-character separators.
fn is_separator(current: char, next: Option<char>) -> bool {
    current.is_ascii_whitespace()
        || matches!(current, '(' | ')' | '[' | ']' | '{' | '}' | ';')
        || (current == '#' && next.is_some_and(|c| matches!(c, '|' | '{')))
}

/// A lexer, used by the parser.
#[derive(Debug)]
pub(crate) struct Lexer<'lexer> {
    input: Chars<'lexer>,
    byte: usize,
    source: Arc<Source>,
}

impl<'lexer> Lexer<'lexer> {
    /// Create a new lexer instance from a string.
    #[must_use]
    pub(crate) fn new(input: &'lexer str) -> Self {
        let source = Arc::new(Source::new(None, input.to_string()));

        Self {
            input: input.chars(),
            byte: 0,
            source,
        }
    }

    /// Set the name of the [Source] being lexically analyzed.
    pub(crate) fn set_name(&mut self, name: String) {
        // TODO: Avoid unwrapping here (if possible?)
        Arc::get_mut(&mut self.source).unwrap().set_name(name);
    }

    /// The source being lexically analyzed.
    #[cfg(test)]
    #[must_use]
    pub(crate) fn source(&self) -> Arc<Source> {
        self.source.clone()
    }

    /// Get the current position of the lexer.
    #[must_use]
    pub(crate) fn span(&self) -> Span {
        Span::new(self.byte..self.byte, self.source.clone())
    }

    /// Returns `true` when at the end of the input.
    #[must_use]
    pub(crate) fn eof(&self) -> bool {
        self.peek(0).is_none()
    }

    /// Get the current character.
    #[must_use]
    fn current(&self) -> Option<char> {
        self.input.as_str().chars().next()
    }

    /// Get the nth character ahead of the current character without advancing.
    #[must_use]
    fn peek(&self, n: usize) -> Option<char> {
        self.input.as_str().chars().nth(n)
    }

    /// Advance the lexer by one character.
    fn advance(&mut self) -> Option<char> {
        let c = self.input.next()?;
        self.byte += c.len_utf8();

        Some(c)
    }

    /// Advance the lexer by one character, and then return the specified
    /// `TokenKind`:
    #[must_use]
    fn advance_and(&mut self, kind: TokenKind) -> TokenKind {
        self.advance();

        kind
    }

    /// Read a word from the input until a separator is reached.
    fn read_word(&mut self) -> String {
        let mut word = String::new();
        while let Some(c) = self.current() {
            if is_separator(c, self.peek(1)) {
                break;
            }

            word.push(c);
            self.advance();
        }

        word
    }

    /// Read the next token from the input.
    #[must_use]
    pub(crate) fn read(&mut self) -> Result<Option<Token>, LexerError> {
        // Eat whitespace until we encounter a meaningful character, or simply return if
        // we have reached the end of input and no additional characters can be read:
        let c = loop {
            match self.current() {
                Some(c) if c.is_ascii_whitespace() => {
                    self.advance();
                }
                Some(c) => break c,
                None => return Ok(None),
            }
        };

        let mut span = self.span();

        let kind = match c {
            ';' => self.line_comment(),
            '#' if self.peek(1) == Some('|') => self.block_comment(),

            '(' => self.advance_and(TokenKind::OpenParen),
            ')' => self.advance_and(TokenKind::CloseParen),
            '{' => self.advance_and(TokenKind::OpenBrace),
            '}' => self.advance_and(TokenKind::CloseBrace),
            '[' => self.advance_and(TokenKind::OpenBracket),
            ']' => self.advance_and(TokenKind::CloseBracket),
            '#' if self.peek(1) == Some('{') => {
                self.advance(); // '#'
                self.advance(); // '{'

                TokenKind::OpenHashBrace
            }

            '0' if matches!(self.peek(1), Some('b') | Some('o') | Some('x')) => {
                let radix = match self.peek(1) {
                    Some('b') => 2,
                    Some('o') => 8,
                    Some('x') => 16,
                    _ => unreachable!(),
                };

                let span = span.clone();
                let word = self.read_word();

                self.integer_literal(word, span, radix)?
            }
            '0'..='9' => self.numeric_literal(span.clone())?,
            '+' | '-' if matches!(self.peek(1), Some('0'..='9')) => {
                self.numeric_literal(span.clone())?
            }

            '\'' => self.char_literal(span.clone())?,
            '"' => self.string_literal(span.clone())?,

            ':' => {
                self.advance(); // ':'

                TokenKind::Keyword(Symbol::from(self.read_word()))
            }

            _ => {
                let word = self.read_word();
                match word.as_str() {
                    "true" => TokenKind::Bool(true),
                    "false" => TokenKind::Bool(false),
                    "nil" => TokenKind::Nil,
                    _ => TokenKind::Symbol(Symbol::from(word)),
                }
            }
        };

        span.extend(&self.span());

        Ok(Some(Token::new(kind, span)))
    }

    #[must_use]
    fn line_comment(&mut self) -> TokenKind {
        // Line comments may start with any number of semicolons, so consume however
        // many are present at the beginning of the comment:
        while self.current().is_some_and(|c| c == ';') {
            self.advance();
        }

        // Line comments continue until a newline character is encountered:
        let mut comment = String::new();
        while let Some(c) = self.advance() {
            if c == '\n' {
                break;
            }

            comment.push(c);
        }

        TokenKind::LineComment(comment.trim().into())
    }

    #[must_use]
    fn block_comment(&mut self) -> TokenKind {
        // TODO: This currently allows for unclosed block comments; do we care?
        self.advance(); // '#'
        self.advance(); // '|'

        let mut comment = String::new();
        while let Some(c) = self.advance() {
            if c == '|' && matches!(self.peek(0), Some('#')) {
                self.advance(); // '#'
                break;
            }

            comment.push(c);
        }

        TokenKind::BlockComment(comment.trim().into())
    }

    #[must_use]
    fn float_literal(&self, word: String, span: Span) -> Result<TokenKind, LexerError> {
        let float = word.parse().map_err(|_| {
            LexerError::new(LexerErrorKind::InvalidNumber(word), span.join(&self.span()))
        })?;

        Ok(TokenKind::Float(float))
    }

    #[must_use]
    fn integer_literal(
        &self,
        word: String,
        span: Span,
        radix: u32,
    ) -> Result<TokenKind, LexerError> {
        // For numbers which are not base-10, strip the prefix (e.g. '0b', '0o', '0x'):
        let word = if radix == 10 {
            word
        } else {
            word[2..].to_string()
        };

        let integer = i64::from_str_radix(&word, radix).map_err(|_| {
            LexerError::new(LexerErrorKind::InvalidNumber(word), span.join(&self.span()))
        })?;

        Ok(TokenKind::Integer(integer))
    }

    #[must_use]
    fn numeric_literal(&mut self, span: Span) -> Result<TokenKind, LexerError> {
        let word = self.read_word();

        let kind = if word.contains('.') {
            self.float_literal(word, span)?
        } else {
            self.integer_literal(word, span, 10)?
        };

        Ok(kind)
    }

    #[must_use]
    fn char_literal(&mut self, span: Span) -> Result<TokenKind, LexerError> {
        self.advance(); // '\''

        let c = match self.advance() {
            Some('\\') => match self.advance() {
                Some(c @ ('"' | '\\')) => c,
                Some('n') => '\n',
                Some('r') => '\r',
                Some('t') => '\t',
                Some('e') => '\x1b',
                Some(c) => {
                    self.read_word(); // Recover from the error
                    return Err(LexerError::new(
                        LexerErrorKind::InvalidEscape(c),
                        span.join(&self.span()),
                    ));
                }
                None => {
                    return Err(LexerError::new(
                        LexerErrorKind::UnclosedChar,
                        span.join(&self.span()),
                    ));
                }
            },
            Some(c) => c,
            None => {
                return Err(LexerError::new(
                    LexerErrorKind::UnclosedChar,
                    span.join(&self.span()),
                ))
            }
        };

        if self.advance() != Some('\'') {
            self.read_word(); // Recover from the error
            return Err(LexerError::new(
                LexerErrorKind::UnclosedChar,
                span.join(&self.span()),
            ));
        }

        Ok(TokenKind::Char(c))
    }

    #[must_use]
    fn string_literal(&mut self, span: Span) -> Result<TokenKind, LexerError> {
        self.advance(); // '"'

        let quote_span = span.clone().join(&self.span());
        let mut string = String::new();

        loop {
            let ch_span = self.span();
            string.push(match self.advance() {
                Some('"') => break,
                Some('\\') => match self.advance() {
                    Some(c @ ('"' | '\\')) => c,
                    Some('n') => '\n',
                    Some('r') => '\r',
                    Some('t') => '\t',
                    Some('e') => '\x1b',
                    Some(c) => {
                        self.read_word(); // Recover from the error
                        return Err(LexerError::new(
                            LexerErrorKind::InvalidEscape(c),
                            ch_span.join(&self.span()),
                        ));
                    }
                    None => {
                        return Err(LexerError::new(LexerErrorKind::UnclosedString, quote_span))
                    }
                },
                Some(c) => c,
                None => return Err(LexerError::new(LexerErrorKind::UnclosedString, quote_span)),
            });
        }

        if self
            .current()
            .is_some_and(|c| !is_separator(c, self.peek(1)))
        {
            self.read_word(); // Recover from the error
            return Err(LexerError::new(
                LexerErrorKind::InvalidString,
                span.join(&self.span()),
            ));
        }

        Ok(TokenKind::String(string))
    }
}

impl Iterator for Lexer<'_> {
    type Item = Result<Token, LexerError>;

    fn next(&mut self) -> Option<Self::Item> {
        self.read().transpose()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    macro_rules! test {
        ( $name:ident: $input:literal, $tokens:expr ) => {
            #[test]
            fn $name() {
                let mut lexer = Lexer::new($input);
                for token in $tokens {
                    let kind = lexer.next().map(|r| match r {
                        Ok(t) => Ok(t.kind),
                        Err(e) => Err(e.kind),
                    });
                    assert_eq!(kind, Some(token));
                }
                assert_eq!(lexer.next(), None);
            }
        };
    }

    test!(block_comment: "foo#| bar\nbaz |#qux", [
        Ok(TokenKind::Symbol(Symbol::from("foo"))),
        Ok(TokenKind::BlockComment("bar\nbaz".into())),
        Ok(TokenKind::Symbol(Symbol::from("qux"))),
    ]);

    test!(line_comment: "nil ;; foo; bar\nnil", [
        Ok(TokenKind::Nil),
        Ok(TokenKind::LineComment("foo; bar".into())),
        Ok(TokenKind::Nil),
    ]);

    test!(list: "(1 () -2.3)", [
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Integer(1)),
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::CloseParen),
        Ok(TokenKind::Float(-2.3)),
        Ok(TokenKind::CloseParen),
    ]);

    test!(set: "{{} true false}", [
        Ok(TokenKind::OpenBrace),
        Ok(TokenKind::OpenBrace),
        Ok(TokenKind::CloseBrace),
        Ok(TokenKind::Bool(true)),
        Ok(TokenKind::Bool(false)),
        Ok(TokenKind::CloseBrace),
    ]);

    test!(vector: "[0 10 200]", [
        Ok(TokenKind::OpenBracket),
        Ok(TokenKind::Integer(0)),
        Ok(TokenKind::Integer(10)),
        Ok(TokenKind::Integer(200)),
        Ok(TokenKind::CloseBracket),
    ]);

    test!(map: "#{:a 0 :b 1}", [
        Ok(TokenKind::OpenHashBrace),
        Ok(TokenKind::Keyword(Symbol::from("a"))),
        Ok(TokenKind::Integer(0)),
        Ok(TokenKind::Keyword(Symbol::from("b"))),
        Ok(TokenKind::Integer(1)),
        Ok(TokenKind::CloseBrace),
    ]);

    test!(number: "0 -1 20.0 +0.003", [
        Ok(TokenKind::Integer(0)),
        Ok(TokenKind::Integer(-1)),
        Ok(TokenKind::Float(20.0)),
        Ok(TokenKind::Float(0.003)),
    ]);

    test!(number_non_base_10: "0b0011 0o70 0xFF", [
        Ok(TokenKind::Integer(3)),
        Ok(TokenKind::Integer(56)),
        Ok(TokenKind::Integer(255)),
    ]);

    test!(error_parse_number: "1.1.1 0.x 7b", [
        Err(LexerErrorKind::InvalidNumber("1.1.1".into())),
        Err(LexerErrorKind::InvalidNumber("0.x".into())),
        Err(LexerErrorKind::InvalidNumber("7b".into())),
    ]);

    test!(char_literal: r"'x' '\n' '\r' '\t' '\e' '\\' '\q' 'b", [
        Ok(TokenKind::Char('x')),
        Ok(TokenKind::Char('\n')),
        Ok(TokenKind::Char('\r')),
        Ok(TokenKind::Char('\t')),
        Ok(TokenKind::Char('\x1b')),
        Ok(TokenKind::Char('\\')),
        Err(LexerErrorKind::InvalidEscape('q')),
        Err(LexerErrorKind::UnclosedChar),
    ]);

    test!(error_unclosed_char_escape: r"'\", [
        Err(LexerErrorKind::UnclosedChar),
    ]);

    test!(error_unclosed_char_empty: r"'", [
        Err(LexerErrorKind::UnclosedChar),
    ]);

    test!(string: "\"\" \"xyz\" \"This is a string!\" \"凄い😍\"", [
        Ok(TokenKind::String("".into())),
        Ok(TokenKind::String("xyz".into())),
        Ok(TokenKind::String("This is a string!".into())),
        Ok(TokenKind::String("凄い😍".into())),
    ]);

    test!(string_with_escapes: "\"\\e[0mfoo\\\"\\nbar\\r\\t\"", [
        Ok(TokenKind::String("\x1b[0mfoo\"\nbar\r\t".into())),
    ]);

    test!(error_invalid_escape_string: "\"oh no \\p\"", [
        Err(LexerErrorKind::InvalidEscape('p')),
    ]);

    test!(error_unclosed_string: "\"hiii", [
        Err(LexerErrorKind::UnclosedString),
    ]);

    test!(error_escape_unclosed_string: "\"oops\\", [
        Err(LexerErrorKind::UnclosedString),
    ]);

    test!(error_invalid_string: "\"hiii\"222", [
        Err(LexerErrorKind::InvalidString),
    ]);

    test!(nested_lists: "(+ (- 0 -1)\n(* 2 3))", [
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("+"))),
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("-"))),
        Ok(TokenKind::Integer(0)),
        Ok(TokenKind::Integer(-1)),
        Ok(TokenKind::CloseParen),
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("*"))),
        Ok(TokenKind::Integer(2)),
        Ok(TokenKind::Integer(3)),
        Ok(TokenKind::CloseParen),
        Ok(TokenKind::CloseParen),
    ]);

    test!(unicode_symbol: "(かわいい 🐕 😻)", [
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("かわいい"))),
        Ok(TokenKind::Symbol(Symbol::from("🐕"))),
        Ok(TokenKind::Symbol(Symbol::from("😻"))),
        Ok(TokenKind::CloseParen),
    ]);
}
